@@ -3,7 +3,7 @@ use strict;
 use warnings;
 package YAML::PP::Parser;
 
-our $VERSION = '0.009'; # VERSION
+our $VERSION = '0.010'; # VERSION
 
 use constant TRACE => $ENV{YAML_PP_TRACE} ? 1 : 0;
 use constant DEBUG => ($ENV{YAML_PP_DEBUG} || $ENV{YAML_PP_TRACE}) ? 1 : 0;
@@ -80,7 +80,7 @@ sub init {
     my ($self) = @_;
     $self->set_offset([]);
     $self->set_events([]);
-    $self->set_new_node(undef);
+    $self->set_new_node(0);
     $self->set_tagmap({
         '!!' => "tag:yaml.org,2002:",
     });
@@ -92,13 +92,13 @@ sub init {
 
 sub parse_string {
     my ($self, $yaml) = @_;
-    $self->lexer->reader->set_input( $yaml );
+    $self->set_reader(YAML::PP::Reader->new( input => $yaml ));
     $self->parse();
 }
 
 sub parse_file {
     my ($self, $file) = @_;
-    $self->lexer->reader->set_input( $file );
+    $self->set_reader(YAML::PP::Reader::File->new( input => $file ));
     $self->parse();
 }
 
@@ -249,6 +249,7 @@ my %nodetypes = (
     FLOWMAP => 'NODETYPE_FLOWMAP',
     FLOWMAPVALUE => 'NODETYPE_FLOWMAPVALUE',
     FLOWSEQ => 'NODETYPE_FLOWSEQ',
+    FLOWSEQ_NEXT => 'FLOWSEQ_NEXT',
 );
 
 sub parse_document {
@@ -259,14 +260,16 @@ sub parse_document {
     my $next_tokens = $lexer->next_tokens;
     my $event_types = $self->events;
     my $stack = $self->event_stack;
-    $lexer->fetch_next_tokens(0);
     LINE: while (1) {
 
         TRACE and $self->info("----------------------- LOOP");
         TRACE and $self->debug_events;
 
         unless (@$next_tokens) {
-            return $self->end_document;
+            $lexer->fetch_next_tokens(0);
+            unless (@$next_tokens) {
+                return $self->end_document;
+            }
         }
         if ( $next_tokens->[0]->{name} eq 'EOL' ) {
             push @{ $self->tokens }, shift @$next_tokens;
@@ -280,22 +283,15 @@ sub parse_document {
         }
 
         DEBUG and $self->info("----------------> parse_next_line");
-        while (1) {
-            unless ($self->new_node) {
-                $self->set_rule( $nodetypes{ $event_types->[-1] } );
-            }
 
-            my $res = $self->parse_tokens();
-
-#            if (@$stack) {
-#                $self->process_events( $res );
-#            }
-
-            last if (not @$next_tokens or $next_tokens->[0]->{column} == 0);
-        }
-        unless (@$next_tokens) {
-            $lexer->fetch_next_tokens(0);
+        unless ($self->new_node) {
+            my $new_rule = $nodetypes{ $event_types->[-1] }
+                or die "Expected document end";
+            $self->set_rule( $new_rule );
         }
+
+        my $res = $self->parse_tokens();
+
     }
 
     return;
@@ -328,13 +324,16 @@ sub check_indent {
     elsif ($self->level < 2 and not $self->new_node) {
         return 1;
     }
-    if ($event_types->[-1] =~ m/^FLOW/) {
-        return;
-    }
 
     my $indent = $self->offset->[ -1 ];
 
     TRACE and $self->info("INDENT: space=$space indent=$indent");
+    if ($event_types->[-1] =~ m/^FLOW/) {
+        if ($space < $indent) {
+            $self->exception("Bad indendation in " . $self->events->[-1]);
+        }
+        return;
+    }
 
     if ($space > $indent) {
         unless ($self->new_node) {
@@ -395,19 +394,12 @@ my %next_event = (
     SEQ0 => 'SEQ0',
     DOC => 'DOC',
     STR => 'STR',
-    FLOWSEQ => 'FLOWSEQ',
+    FLOWSEQ => 'FLOWSEQ_NEXT',
+    FLOWSEQ_NEXT => 'FLOWSEQ',
     FLOWMAP => 'FLOWMAPVALUE',
     FLOWMAPVALUE => 'FLOWMAP',
 );
 
-my %render_methods = (
-    q/:/ => 'render_multi_val',
-    q/"/ => 'render_quoted',
-    q/'/ => 'render_quoted',
-    q/>/ => 'render_block_scalar',
-    q/|/ => 'render_block_scalar',
-);
-
 my %event_to_method = (
     MAP => 'mapping',
     FLOWMAP => 'mapping',
@@ -477,36 +469,51 @@ my %event_to_method = (
 #    }
 #}
 
+my %fetch_method = (
+    '"' => 'fetch_quoted',
+    "'" => 'fetch_quoted',
+    '|' => 'fetch_block',
+    '>' => 'fetch_block',
+    ''  => 'fetch_plain',
+);
+
 sub parse_tokens {
     my ($self) = @_;
     my $res = {};
     my $next_rule_name = $self->rule;
     DEBUG and $self->info("----------------> parse_tokens($next_rule_name)");
-    my $next_rule = $GRAMMAR->{ $next_rule_name };
+    my $next_rule = $GRAMMAR->{ $next_rule_name }
+        or die "Could not find rule $next_rule_name";
 
     TRACE and $self->debug_rules($next_rule);
     TRACE and $self->debug_yaml;
     DEBUG and $self->debug_next_line;
 
+    my $event_types = $self->events;
+    my $offsets = $self->offset;
     my $tokens = $self->tokens;
     my $next_tokens = $self->lexer->next_tokens;
     $res->{offset} = $next_tokens->[0]->{column};
     RULE: while ($next_rule_name) {
         DEBUG and $self->info("RULE: $next_rule_name");
+        TRACE and $self->debug_tokens($next_tokens);
 
         unless (@$next_tokens) {
             $self->exception("No more tokens");
         }
         TRACE and warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\$next_tokens->[0]], ['next_token']);
         my $got = $next_tokens->[0]->{name};
+        if ($got eq 'CONTEXT') {
+            my $context = shift @$next_tokens;
+            my $indent = $offsets->[-1];
+            $indent++ unless $self->lexer->flowcontext;
+            my $method = $fetch_method{ $context->{value} };
+            my $partial = $self->lexer->$method($indent, $context->{value});
+            next RULE;
+        }
         my $def = $next_rule->{ $got };
         if ($def) {
-            if ($got eq 'END') {
-                shift @$next_tokens;
-            }
-            else {
-                push @$tokens, shift @$next_tokens;
-            }
+            push @$tokens, shift @$next_tokens;
         }
         elsif ($def = $next_rule->{DEFAULT}) {
             $got = 'DEFAULT';
@@ -523,30 +530,35 @@ sub parse_tokens {
             DEBUG and $self->info("CALLBACK $sub");
             $self->$sub($tokens->[-1]);
         }
-        my $node = $def->{node};
-        my $new = $node || $def->{new};
+        my $new = $def->{new};
         if ($new) {
+            DEBUG and $self->got("NEW: $new");
+
             $next_rule_name = $new;
-            DEBUG and $self->got("NEW: $next_rule_name");
-#            if ($node) {
-#                $self->set_new_node($node);
-#            }
+            $self->set_rule($next_rule_name);
 
             if ($def->{return}) {
-                $self->set_rule($next_rule_name);
-                return $res;
+                return 1;
             }
 
-            $next_rule = $GRAMMAR->{ $next_rule_name }
-                or die "Unexpected rule $next_rule_name";
-            next RULE;
         }
         elsif ($def->{return}) {
-            $self->set_new_node(undef);
-            return $res;
+            if ($got eq 'EOL' or not @$next_tokens) {
+                return 1;
+            }
+
+            $next_rule_name = $nodetypes{ $event_types->[-1] }
+                or die "Expected document end";
+            $self->set_rule($next_rule_name);
+
         }
-        $next_rule_name .= " - $got"; # for debugging
-        $next_rule = $def;
+        else {
+            $next_rule_name .= " - $got"; # for debugging
+            $next_rule = $def;
+            next RULE;
+        }
+        $next_rule = $GRAMMAR->{ $next_rule_name }
+            or die "Unexpected rule $next_rule_name";
 
     }
 
@@ -636,8 +648,17 @@ sub start_sequence {
 sub start_flow_sequence {
     my ($self, $offset) = @_;
     my $offsets = $self->offset;
+    my $new_offset = $offsets->[-1];
+    my $event_types = $self->events;
+    if ($new_offset < 0) {
+        $new_offset = 0;
+    }
+    elsif ($self->new_node) {
+        $new_offset++;
+    }
     push @{ $self->events }, 'FLOWSEQ';
-    push @{ $offsets }, $offset;
+    push @{ $offsets }, $new_offset;
+
     my $event_stack = $self->event_stack;
     my $info = { style => 'flow' };
     if (@$event_stack and $event_stack->[-1]->[0] eq 'properties') {
@@ -649,8 +670,17 @@ sub start_flow_sequence {
 sub start_flow_mapping {
     my ($self, $offset) = @_;
     my $offsets = $self->offset;
+    my $new_offset = $offsets->[-1];
+    my $event_types = $self->events;
+    if ($new_offset < 0) {
+        $new_offset = 0;
+    }
+    elsif ($self->new_node) {
+        $new_offset++;
+    }
     push @{ $self->events }, 'FLOWMAP';
-    push @{ $offsets }, $offset;
+    push @{ $offsets }, $new_offset;
+
     my $event_stack = $self->event_stack;
     my $info = { style => 'flow' };
     if (@$event_stack and $event_stack->[-1]->[0] eq 'properties') {
@@ -759,11 +789,13 @@ sub scalar_event {
         my $properties = pop @$event_stack;
         $properties = $self->node_properties($properties->[1], $info);
     }
-    my $method = $render_methods{ $info->{style} };
-    YAML::PP::Render->$method( $info );
 
     $self->callback->($self, 'scalar_event', $info);
-    $self->set_new_node(undef);
+    if ($event_types->[-1] =~ m/^FLOW/) {
+    }
+    else {
+        $self->set_new_node(0);
+    }
     $event_types->[-1] = $next_event{ $event_types->[-1] };
 }
 
@@ -775,7 +807,11 @@ sub alias_event {
     }
     my $event_types = $self->events;
     $self->callback->($self, 'alias_event', $info);
-    $self->set_new_node(undef);
+    if ($event_types->[-1] =~ m/^FLOW/) {
+    }
+    else {
+        $self->set_new_node(0);
+    }
     $event_types->[-1] = $next_event{ $event_types->[-1] };
 }
 
@@ -788,19 +824,34 @@ sub yaml_to_tokens {
     my $error = $@;
 
     my $tokens = $yp->tokens;
-    my $next = $yp->lexer->next_tokens;
     if ($error) {
-        push @$tokens, map { +{ %$_, name => 'ERROR' } } @$next;
-        my $remaining = $yp->reader->read;
-        $remaining = '' unless defined $remaining;
-        push @$tokens, { name => "ERROR", value => $remaining };
-    }
-    else {
-        push @$tokens, @$next;
+        my $remaining_tokens = $yp->_remaining_tokens;
+        push @$tokens, map { +{ %$_, name => 'ERROR' } } @$remaining_tokens;
     }
     return $error, $tokens;
 }
 
+sub _remaining_tokens {
+    my ($self) = @_;
+    my @tokens;
+    my $next = $self->lexer->next_tokens;
+    push @tokens, @$next;
+    my $next_line = $self->lexer->next_line;
+    my $remaining = '';
+    if ($next_line) {
+        if ($self->lexer->offset > 0) {
+            $remaining = $next_line->[1] . $next_line->[2];
+        }
+        else {
+            $remaining = join '', @$next_line;
+        }
+    }
+    $remaining .= $self->reader->read;
+    $remaining = '' unless defined $remaining;
+    push @tokens, { name => "ERROR", value => $remaining };
+    return \@tokens;
+}
+
 sub event_to_test_suite {
     my ($self, $event) = @_;
     if (ref $event) {
@@ -1000,16 +1051,28 @@ sub exception {
     my ($self, $msg, %args) = @_;
     my $next = $self->lexer->next_tokens;
     my $line = @$next ? $next->[0]->{line} : $self->lexer->line;
+    my $offset = @$next ? $next->[0]->{column} : $self->lexer->offset;
+    $offset++;
     my $next_line = $self->lexer->next_line;
+    my $remaining = '';
+    if ($next_line) {
+        if ($self->lexer->offset > 0) {
+            $remaining = $next_line->[1] . $next_line->[2];
+        }
+        else {
+            $remaining = join '', @$next_line;
+        }
+    }
     my $caller = $args{caller} || [ caller(0) ];
     my $e = YAML::PP::Exception->new(
         got => $args{got},
         expected => $args{expected},
         line => $line,
+        column => $offset,
         msg => $msg,
         next => $next,
         where => $caller->[1] . ' line ' . $caller->[2],
-        yaml => $next_line,
+        yaml => $remaining,
     );
     croak $e;
 }
@@ -1094,6 +1157,7 @@ sub cb_send_scalar {
     my ($self, $res) = @_;
     my $last = pop @{ $self->event_stack };
     $self->scalar_event($last->[1]);
+    $self->set_new_node(0);
 }
 
 sub cb_empty_mapkey {
@@ -1111,18 +1175,24 @@ sub cb_empty_mapkey {
     $self->set_new_node(1);
 }
 
+sub cb_send_flow_alias {
+    my ($self, $token) = @_;
+    my $alias = substr($token->{value}, 1);
+    $self->alias_event({ value => $alias });
+}
+
 sub cb_send_alias {
     my ($self, $token) = @_;
     my $alias = substr($token->{value}, 1);
     $self->alias_event({ value => $alias });
-    $self->set_new_node(1);
+    $self->set_new_node(0);
 }
 
 sub cb_send_alias_from_stack {
     my ($self, $token) = @_;
     my $last = pop @{ $self->event_stack };
     $self->alias_event($last->[1]);
-    $self->set_new_node(1);
+    $self->set_new_node(0);
 }
 
 sub cb_alias {
@@ -1139,6 +1209,10 @@ sub cb_question {
     $self->set_new_node(1);
 }
 
+sub cb_flow_question {
+    my ($self, $res) = @_;
+}
+
 sub cb_empty_complexvalue {
     my ($self, $res) = @_;
     $self->scalar_event({ style => ':', value => undef });
@@ -1166,12 +1240,13 @@ sub cb_seqitem {
     $self->set_new_node(1);
 }
 
-sub cb_start_quoted {
+sub cb_take_quoted {
     my ($self, $token) = @_;
+    my $subtokens = $token->{subtokens};
     my $stack = $self->event_stack;
     my $info = {
-        style => $token->{value},
-        value => [],
+        style => $subtokens->[0]->{value},
+        value => $token->{value},
         offset => $token->{column},
     };
     if (@$stack and $stack->[-1]->[0] eq 'properties') {
@@ -1180,10 +1255,41 @@ sub cb_start_quoted {
     push @{ $stack }, [ scalar => $info ];
 }
 
-sub cb_fetch_tokens_quoted {
-    my ($self) = @_;
-    my $indent = $self->offset->[-1] + 1;
-    $self->lexer->fetch_next_tokens($indent);
+sub cb_quoted_multiline {
+    my ($self, $token) = @_;
+    my $subtokens = $token->{subtokens};
+    my $stack = $self->event_stack;
+    my $info = {
+        style => $subtokens->[0]->{value},
+        value => $token->{value},
+        offset => $token->{column},
+    };
+    if (@$stack and $stack->[-1]->[0] eq 'properties') {
+        $self->fetch_inline_properties($stack, $info);
+    }
+    push @{ $stack }, [ scalar => $info ];
+    $self->cb_send_scalar;
+}
+
+sub cb_take_quoted_key {
+    my ($self, $token) = @_;
+    $self->cb_take_quoted($token);
+    $self->cb_send_mapkey;
+}
+
+sub cb_send_plain_multi {
+    my ($self, $token) = @_;
+    my $stack = $self->event_stack;
+    my $info = {
+        style => ':',
+        value => $token->{value},
+        offset => $token->{column},
+    };
+    if (@$stack and $stack->[-1]->[0] eq 'properties') {
+        $self->fetch_inline_properties($stack, $info);
+    }
+    push @{ $stack }, [ scalar => $info ];
+    $self->cb_send_scalar;
 }
 
 sub cb_start_plain {
@@ -1191,7 +1297,7 @@ sub cb_start_plain {
     my $stack = $self->event_stack;
     my $info = {
             style => ':',
-            value => [ $token->{value} ],
+            value => $token->{value},
             offset => $token->{column},
     };
     if (@$stack and $stack->[-1]->[0] eq 'properties') {
@@ -1200,41 +1306,29 @@ sub cb_start_plain {
     push @{ $stack }, [ scalar => $info ];
 }
 
-sub cb_empty_plain {
-    my ($self, $res) = @_;
-    my $stack = $self->event_stack;
-    push @{ $stack->[-1]->[1]->{value} }, '';
-    $self->cb_fetch_tokens_plain;
-}
-
-sub cb_fetch_tokens_plain {
-    my ($self) = @_;
-    my $indent = $self->offset->[-1] + 1;
-    $self->lexer->set_context('plain');
-    $self->lexer->fetch_next_tokens($indent);
-}
-
 sub cb_start_flowseq {
     my ($self, $token) = @_;
     $self->start_flow_sequence($token->{column});
-    $self->set_new_node(1);
 }
 
 sub cb_start_flowmap {
     my ($self, $token) = @_;
     $self->start_flow_mapping($token->{column});
-    $self->set_new_node(1);
 }
 
 sub cb_end_flowseq {
     my ($self, $res) = @_;
     $self->end_flow_sequence;
-    $self->set_new_node(undef);
+    $self->set_new_node(0);
 }
 
 sub cb_flow_comma {
     my ($self) = @_;
-    $self->set_new_node(1);
+    my $event_types = $self->events;
+    $self->set_new_node(0);
+    if ($event_types->[-1] =~ m/^FLOWSEQ/) {
+        $event_types->[-1] = $next_event{ $event_types->[-1] };
+    }
 }
 
 sub cb_flow_colon {
@@ -1242,10 +1336,77 @@ sub cb_flow_colon {
     $self->set_new_node(1);
 }
 
+sub cb_empty_flow_mapkey {
+    my ($self, $token) = @_;
+    my $stack = $self->event_stack;
+    my $info = {
+        style => ':',
+        value => undef,
+        offset => $token->{column},
+    };
+    if (@$stack and $stack->[-1]->[0] eq 'properties') {
+        $self->fetch_inline_properties($stack, $info);
+    }
+    $self->scalar_event($info);
+    $self->set_new_node(0);
+}
+
 sub cb_end_flowmap {
     my ($self, $res) = @_;
     $self->end_flow_mapping;
-    $self->set_new_node(undef);
+    $self->set_new_node(0);
+}
+
+sub cb_end_flowmap_empty {
+    my ($self, $res) = @_;
+    $self->cb_empty_flowmap_value;
+    $self->end_flow_mapping;
+    $self->set_new_node(0);
+}
+
+sub cb_flow_plain {
+    my ($self, $token) = @_;
+    my $stack = $self->event_stack;
+    my $info = {
+        style => ':',
+        value => $token->{value},
+        offset => $token->{column},
+    };
+    if (@$stack and $stack->[-1]->[0] eq 'properties') {
+        $self->fetch_inline_properties($stack, $info);
+    }
+    $self->scalar_event($info);
+}
+
+sub cb_flowkey_plain {
+    my ($self, $token) = @_;
+    my $stack = $self->event_stack;
+    my $info = {
+        style => ':',
+        value => $token->{value},
+        offset => $token->{column},
+    };
+    if (@$stack and $stack->[-1]->[0] eq 'properties') {
+        $self->fetch_inline_properties($stack, $info);
+    }
+    $self->scalar_event($info);
+    $self->set_new_node(0);
+}
+
+sub cb_flowkey_quoted {
+    my ($self, $token) = @_;
+    my $stack = $self->event_stack;
+    my $subtokens = $token->{subtokens};
+    my $info = {
+        style => $subtokens->[0]->{value},
+        value => $token->{value},
+        offset => $token->{column},
+    };
+    if (@$stack and $stack->[-1]->[0] eq 'properties') {
+        $self->fetch_inline_properties($stack, $info);
+    }
+    $self->scalar_event($info);
+    $self->set_new_node(0);
 }
 
 sub cb_empty_flowmap_value {
@@ -1260,20 +1421,6 @@ sub cb_empty_flowmap_value {
         $self->fetch_inline_properties($stack, $info);
     }
     $self->scalar_event($info);
-    $self->set_new_node(undef);
-}
-
-sub cb_take {
-    my ($self, $token) = @_;
-    my $stack = $self->event_stack;
-    push @{ $stack->[-1]->[1]->{value} }, $token->{value};
-}
-
-sub cb_empty_quoted_line {
-    my ($self, $res) = @_;
-    my $stack = $self->event_stack;
-    push @{ $stack->[-1]->[1]->{value} }, '';
-    $self->cb_fetch_tokens_quoted;
 }
 
 sub cb_insert_map_alias {
@@ -1312,69 +1459,23 @@ sub cb_insert_empty_map {
     $self->set_new_node(1);
 }
 
-sub cb_block_scalar {
+sub cb_send_block_scalar {
     my ($self, $token) = @_;
-    my $type = $token->{value};
+    my $type = $token->{subtokens}->[0]->{value};
     my $stack = $self->event_stack;
     my $info = {
         style => $type,
-        value => [],
-        current_indent => $self->offset->[-1] + 1,
+        value => $token->{value},
         offset => $token->{column},
     };
     if (@$stack and $stack->[-1]->[0] eq 'properties') {
         $self->fetch_inline_properties($stack, $info);
     }
     push @{ $self->event_stack }, [ scalar => $info ];
-    $self->lexer->set_context('block_scalar_start');
-}
+    $self->cb_send_scalar;
 
-sub cb_add_block_scalar_indent {
-    my ($self, $token) = @_;
-    my $indent = $token->{value};
-    my $event = $self->event_stack->[-1]->[1];
-    $event->{block_indent} = $indent;
-    $event->{got_indent} = 1;
-    $event->{current_indent} = $indent;
-    $self->lexer->set_context('block_scalar');
-}
-
-sub cb_add_block_scalar_chomp {
-    my ($self, $token) = @_;
-    my $chomp = $token->{value};
-    $self->event_stack->[-1]->[1]->{block_chomp} = $chomp;
-}
-
-sub cb_block_scalar_empty_line {
-    my ($self, $res) = @_;
-    my $event = $self->event_stack->[-1]->[1];
-    push @{ $event->{value} }, '';
-    $self->lexer->fetch_next_tokens($event->{current_indent});
-}
-
-sub cb_block_scalar_start_indent {
-    my ($self, $token) = @_;
-    my $event = $self->event_stack->[-1]->[1];
-    $event->{current_indent} = length $token->{value};
-}
-
-sub cb_fetch_tokens_block_scalar {
-    my ($self, $res) = @_;
-    my $event = $self->event_stack->[-1]->[1];
-    $self->lexer->fetch_next_tokens($event->{current_indent})
-}
-
-sub cb_block_scalar_start_content {
-    my ($self, $token) = @_;
-    my $event = $self->event_stack->[-1]->[1];
-    push @{ $event->{value} }, $token->{value};
-    $self->lexer->set_context('block_scalar');
-}
-
-sub cb_block_scalar_content {
-    my ($self, $token) = @_;
-    my $event = $self->event_stack->[-1]->[1];
-    push @{ $event->{value} }, $token->{value};
+    my $indent = $self->offset->[-1] + 1;
+    $self->lexer->fetch_next_tokens($indent);
 }
 
 1;
